// ForwardDeclarations.h
// 2/27/04
// Forrest Briggs
// Including this file prevents the compiler from complaining 
// about incomplete definitions, and allows each header file
// to include only the other headers it needs.

#ifndef forwarddeclarations_h
#define forwarddeclarations_h

class cTimer;

class cColor;
class cTextureCoord;
class cTextureImage;
class cVector;
class cVertex;
class cTriangle;
class cStandardizedTriangle;
//class cFace;
class cEdge;
class cStaticTriangleMesh;
class cIntersectionResult;
class cBall;
class cWorld;
class cCamera;

class cOctree;
class cBoundingBox;

class cMS3DFileLoader;

class cSnowParticle;
class cSmokeTextParticle;
class cSmokeRingParticle;
class cSplashParticle;
class cHoleArrowParticle;
class cParticleSystem;
class cParticle;

struct cVectorComparator;
struct cEdgeComparator;
#endif